module Sem08 where

-- Импорты, чтобы компилятор не ругался на дубликаты инстансов

import qualified Prelude as P
import qualified Control.Applicative as A


-- Функторы. Обычные(немного) и аппликативные

-- Я мог бы снова сорваться на теоркат, но не буду этого делать(постараюсь)
-- В С++(по модулю того, что я на нем давно не писал) функтор -- это объект-функция - жалкая попытка дать функциям гражданство. В хаскеле все не так
-- О функторе удобно думать как о типе, который можно отобразить.
-- Например если у нас есть список чисел и функция, переводящая число в строку,
-- то благодаря структуре функтора мы можем получать списки,
-- в которых лежат строковые представления чисел. Я утрирую, но общая идея такова

-- Определение класса типов Functor:

class Functor f where
    fmap :: (a -> b) -> f a -> f b

-- У fmap есть инфиксный синоним:

infixl 4 <$>

(<$>) :: Functor f => (a -> b) -> f a -> f b
(<$>) = fmap

-- У функторов есть два закона, которые растут из теоретико-категориального определения
-- и суть которых состоит в том, что при отображении нашего типа его структура не должна меняться.

-- Можно убедиться, что законы функтора выполняются для списков.
-- Формально, можно доказать, что для любого списка xs и подходящих функций f и g верно:
-- 1. fmap id = id
-- 2. fmap (f . g) = fmap f . fmap g

-- Все ли в порядке с утверждением выше? Точно? Если не в порядке, то где кроется подвох?

-- В отличии от обычных, аппликативные функторы позволяют вам не просто применять чистую функцию к значению в контейнере, но и "вкладывать" её в контейнер. У аппликативных функторов так же есть законы, которым должен подчиняться каждый представитель:

-- 1. pure id <*> v = v
-- 2. pure f <*> pure x = pure (f x)
-- 3. pure (.) <*> f <*> g <*> x = f <*> (g <*> x)
-- 4. f <*> pure x = pure ($ x) <*> f

-- Определение класса типов Applicative:

class Functor f => Applicative f where
    pure :: a -> f a
    -- произносится как apply
    (<*>) :: f (a -> b) -> f a -> f b

-- В качестве разминки:
-- Выразите с помощью apply(<*>) метод <*
-- Его тип похож на const, но, в отличии от него, он позволяет "поработать"
-- аппликативу, при этом забив на второй аргумент
-- Например:
-- [1,2,3] `const` [3,4,5] = [1,2,3]
-- [1,2,3] <* [3,4,5] = [1,1,1,2,2,2,3,3,3]
-- Произносится как left apply

infixl 4 <*

(<*) :: Applicative f => f a -> f b -> f a
(<*) fa fb = P.const <$> fa <*> fb

-- В качестве разминки. Чему равны значения выражений:
-- [(*0), (+100), (^2)] <*> [1,2,3]
-- ([(+), (*)] <*> [1,2]) <*> [3,4]
-- getZipList $ (,,) <$> ZipList "dog" <*> ZipList "cat" <*> ZipList "rat"

-- Как мы узнали на лекции, тип функций с фиксированным доменом -- это функтор.
-- где fmap -- это всего-навсего композиция.
-- Напишем для него представителя Applicative

instance Functor ((->) e) where
    fmap f g x = f (g x)

infixl 4 <*>

instance Applicative ((->) e) where
    pure = P.const
    (<*>) f g x = f x (g x)

-- Выпишем тип <*>:
-- (e -> (a -> b)) -> (e -> a) -> (e -> b)
-- Или же(если убрать лишние скобки):
-- (e -> a -> b) -> (e -> a) -> e -> b
-- Ничего не напоминает?

-- Пусть мы все еще рассматриваем наш тип-стрелку. Какой тип имеют f g h?
-- 1. \f g h -> f <*> g <*> h

t1 :: (e -> a -> b -> c) -> (e -> a) -> (e -> b) -> e -> c
t1 = \f g h -> f <*> g <*> h

t1' :: (e -> a -> b -> c) -> (e -> a) -> (e -> b) -> e -> c
t1' = \f g h x -> f x (g x) (h x)


-- 2. \f g h -> (f <$> g) <*> h

-- (f <$> g) <*> h :: e -> c
-- h :: e -> b
-- (f <$> g) :: t (b -> c)
-- g :: e -> a
-- f :: a -> b -> c

t2 :: (a -> b -> c) -> (e -> a) -> (e -> b) -> e -> c
t2 = \f g h -> f <$> g <*> h

t2'' :: (a -> b -> c) -> (e -> a) -> (e -> b) -> e -> c
t2'' = \f g h x -> f (g x) (h x)


-- Как можно переписать эти конструкции без аппликатива?
-- Хинт -- не обязательно даже лезть в определение,
-- достаточно просто решить уравнения на типы

-- Еще несколько мозголомательных примеров.
-- Чему равны значения следующих выражений:

-- (pure 3) "blah"
-- (+) <*> (*3) $ 4
-- zip <*> tail $ [1..10]
-- (+) <$> (+2) <*> (*3) $ 10


-- Пример Applicative(наконец-то что-то написать!) для копроизведения

instance Functor (P.Either e) where
    fmap _ (P.Left e) = P.Left e
    fmap f (P.Right x) = P.Right (f x)

instance Applicative (P.Either e) where
    pure = P.Right
    P.Right f <*> P.Right x = P.Right (f x)
    P.Left er <*> P.Right _ = P.Left er
    P.Right _ <*> P.Left er = P.Left er
    P.Left e1 <*> P.Left _  = P.Left e1


-- Напишите функцию, которая делает следующее:
-- Берет список значений, лежащих в некотором аппликативном контексте
-- Пересобирает этот список так, что значения вынимаются из контекста,
-- а список вкладывается в него

sequence' :: A.Applicative f => [f a] -> f [a]
sequence' [] = A.pure []
sequence' (x : xs) = let
    rec = sequence' xs
    fun = P.fmap (:) x
    in (A.<*>) fun rec

-- Напишите аппликативный аналог функции replicate.
-- Она должна повторять некоторый аппликативный эффект заданое число раз

replicateA' :: A.Applicative f => P.Int -> f a -> f [a]
replicateA' n x = sequence' (P.replicate n x)

-- Наконец, напишите функцию, которая фильтрует список с помощью
-- заданного предиката, но который еще и порождает эффект

filterA :: A.Applicative f => (a -> f P.Bool) -> [a] -> f [a]
filterA _ [] = A.pure []
filterA p (x:xs) = let
    rec = filterA p xs
    fun = P.fmap (\b -> if b then (:) x else P.id) (p x)
    in (A.<*>) fun rec

-- Классические комбинаторы на уровне типов
-- У нас были разные комбинаторы в лямбда-исчислении. Давайте поднимем их на уровень типов
-- Например композиция:

newtype Cmps f g x = Cmps { getCmps :: f (g x) } deriving P.Show

-- Какой кайнд у этого конструктора типов? Приведите пример _типа_, полученного с помощью Cmps
-- и _терма_, имеющего данный тип

-- Пусть есть определение функции:
-- ffmap h = getCmps . fmap h . Cmps
-- Какой тип она имеет и, самое главное, что не нравится компилятору, когда он пытается тайпчекнуть это определение?

-- Попробуйте вызвать ее на чем-нибудь! Если компилятор продолжает отвергать ваш код -- это нормально.
-- Для починки напишите инстанс функтора для нашего оператора над типами

instance (Functor f, Functor g) => Functor (Cmps f g) where
    fmap f (Cmps x) = Cmps (fmap (fmap f) x)

-- Можно доказать, что наше определение функториально, таким образом -- композиция функторов -- это функтор. Аналогичное утверждение верно и для Applicative, Foldable и Traversable. К сожалению, оно неверно для Monad =(

-- Parsec
-- Я не успел показать эту штуку на паре, но вот есть известная библиотека парсер-комбинаторов Parsec. Идея в том, что сложные парсеры можно комбинировать из более простых, используя её аппликативный интерфейс.

-- поставить(на всякий случай, она вроде идет из коробки с платформой):
-- $ cabal update
-- $ cabal install parsec

-- GHCi> import Control.Applicative((<*>), (<*), (*>), (<$>))
-- GHCi> import Text.Parsec
-- GHCi> :set -XFlexibleContexts

-- Функция parse применяет парсер rule ко входной строке text.
-- Вторым параметром она принимает имя файла с текстом, на которое мы забъем
-- В результате у нас получается либо результат парсинга, либо сообщение
-- об ошибке с указанием номера строки и символа в ней и более подробным текстом
-- Например:

-- GHCi> let prs rule text = parse rule "" text
-- GHCi> prs digit "123ABC"
-- Right '1'
-- GHCi> prs letter "123ABC"
-- Left (line 1, column 1):
-- unexpected "1"
-- expecting letter

-- digit и letter это парсеры, которые парсят по одной цифре и букве
-- соответственно. Есть парсеры, которые парсят конкретные строки(string)
-- или символы(char). Например:

-- GHCi> prs (string "ABC") "ABC123"
-- Right "ABC"
-- GHCi> prs (char 'A') "ABC123"
-- Right 'A'
-- GHCi> prs (string "ABC") "123ABC"
-- Left (line 1, column 1):
-- unexpected "1"
-- expecting "ABC"
-- GHCi> prs (char 'A') "123ABC"
-- Left (line 1, column 1):
-- unexpected "1"
-- expecting "A"

-- Если мы захотим распарсить не одну цифру, а число, то мы можем модифицировать
-- данный нам парсер для цифр, так что бы он много раз применялся ко входной
-- строке. Для этого нам помогут комбинаторы many(ноль или более раз)
-- или many1(один или более раз). Например:

-- GHCi> prs (many digit) "123ABC"
-- Right "123"
-- GHCi> prs (many digit) "ABC"
-- Right ""
-- GHCi> prs (many1 digit) "123ABC"
-- Right "123"
-- GHCi> prs (many1 digit) "ABC"
-- Left (line 1, column 1):
-- unexpected "A"
-- expecting digit

-- Парсеры можно комбинировать между собой. Например мы хотим распарсить пару
-- из строчки и числа:
-- GHCi> let pairp = (,) <$> (many1 digit) <*> (many1 letter)
-- GHCi> prs pairp "123ABC"
-- Right ("123","ABC")
-- GHCi> prs pairp "ABC123"
-- Left (line 1, column 1):
-- unexpected "A"
-- expecting digit

-- pairp -- это парсер, который сначала парсит один или более раз цифру,
-- употребляя входную строку, затем парсит один или более раз букву, и затем
-- складывает результат в пару. Еще примеры:
-- GHCi> let p0 =  many1 digit <* many1 letter
-- GHCi> prs p0 "123ABC"
-- Right "123"
-- GHCi> let p1 =  many1 digit *> many1 letter
-- GHCi> prs p1 "123ABC"
-- Right "ABC"

-- Когда мы парсим строчку парсером p0, то мы сначала парсим ее левым парсером,
-- запоминаем результат, парсим остаток строки правым парсером, забиваем на его результат и выдаем
-- что получилось до этого. Для p1 -- ровно наоборот, попарсили строку левым парсером,
-- потом правым, забили на результат левого и вернули, что вышло у правого

-- Больше документации вот здесь:
-- https://hackage.haskell.org/package/parsec-3.1.11/docs/Text-Parsec.html
